#include <vector>
#include "pthread.h"
#include <assert.h>

template<class Token>
class RingBuffer {
   public:
      // constructors and destructors
      RingBuffer(unsigned int size);
      ~RingBuffer();

      // mutators
      void add(Token& token);
      Token get();

   private:
      std::vector<Token> buffer;
      unsigned int filled; // # of objects in the buffer
      unsigned int in; // next token is stored here
      unsigned int out; // next token is taken from here
      pthread_mutex_t mutex;
      pthread_cond_t nonEmpty;
      pthread_cond_t nonFull;
}; // class RingBuffer

template<class Token>
RingBuffer<Token>::RingBuffer(unsigned int size) :
   buffer(size), filled(0), in(0), out(0)
{
   pthread_mutex_init(&mutex, 0);
   pthread_cond_init(&nonEmpty, 0);
   pthread_cond_init(&nonFull, 0);
} // constructor

template<class Token>
RingBuffer<Token>::~RingBuffer() {
   pthread_mutex_destroy(&mutex);
   pthread_cond_destroy(&nonEmpty);
   pthread_cond_destroy(&nonFull);
} // destructor

template<class Token>
void RingBuffer<Token>::add(Token& token) {
   pthread_mutex_lock(&mutex);
   while (filled == buffer.size()) {
      pthread_cond_wait(&nonFull, &mutex);
   }
   assert(filled < buffer.size());
   buffer[in] = token;
   in = (in + 1) % buffer.size();
   ++filled;
   pthread_cond_signal(&nonEmpty);
   pthread_mutex_unlock(&mutex);
} // add

template<class Token>
Token RingBuffer<Token>::get() {
   Token token;

   pthread_mutex_lock(&mutex);
   while (filled == 0) {
      pthread_cond_wait(&nonEmpty, &mutex);
   }
   assert(filled > 0);
   token = buffer[out];
   out = (out + 1) % buffer.size();
   --filled;
   pthread_cond_signal(&nonFull);
   pthread_mutex_unlock(&mutex);

   return token;
} // get

